//******************************************************************************
//*              A Mail Filtering and Control Utility   "MailUtl"
//*                     Copyright (C) 2002-2005  Fumio KAWAMATA
//*
//*    This program is free software; you can redistribute it and/or modify
//*    it under the terms of the GNU General Public License as published by
//*    the Free Software Foundation; either version 2 of the License, or
//*    (at your option) any later version.
//*
//*    This program is distributed in the hope that it will be useful,
//*    but WITHOUT ANY WARRANTY; without even the implied warranty of
//*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//*    GNU General Public License for more details.
//*
//******************************************************************************
program MailUtl;

{$APPTYPE CONSOLE}
{$R MailUtl}

{$DEFINE Japanese}
//* Remove above compiler directive if you do not use Japanese.

uses
  SysUtils, Classes, Windows,

    IdPOP3 in 'c:\indy\protocols\idPOP3.pas',
    IdMessage in 'c:\indy\protocols\idMessage.pas',
    IdEmailAddress in 'c:\indy\protocols\idEmailAddress.pas',
    IdException in 'c:\indy\system\idException.pas',
    IdGlobal in 'c:\indy\system\idGlobal.pas',
    BRegExp in 'BregExp\BregExp.pas',
    //* BregExp.pas Copyright (C) Osamu Takeuchi All Rights Reserved.


  {$IFDEF Japanese}
    PZConv in 'PZConv\PZConv.pas',
    //* PZConv.pas Copyright (C) 1998 Pz factory All Rights Reserved.
  {$ENDIF Japanese}
  IniFiles;



const
  Version = '1.42';  //* Revise this before new release.
  PathDelimiter = '\';

type
  TMyParam = record
    Bcc:             string;
    Cc:              string;
    HTML_Lang:       string;
    HTML_Charset:    string;
    InhibitAction:   Boolean;
    IniName:         string;
    LogDir:          string;
    Mailer:          string;
    POP3_Host:       string;
    POP3_Port:       Integer;
    POP3_UserID:     string;
    POP3_Password:   string;
    ReplyTo:         string;
    Subject:         string;
    TemplateDir:     string;
  end;

  TMyAction = (actKeep, actDel, actNop);

  TMyEvaluated= record
    Action:   TMyAction;
    Reason:   string;
  end;


{$IFDEF Japanese}
var
  Jap: TPZConv;
{$ENDIF Japanese}
var
  Field1List: TStringList;
  Field2List: TStringList;
  Field3List: TStringList;
  Param:      TMyParam;


//*****************************************************************************
//* MyStrToInt
//*****************************************************************************
function MyStrToInt(S: string): Integer;
begin
  try
    Result:=StrToInt(S);
  except
    Result:=0;
  end;
end;


//*****************************************************************************
//* EnterPassword
//*   Entering password
//*****************************************************************************
function EnterPassword(Item: string): string;
var
  S: string;
begin
  Write('Enter '+Item+' password: ');
  Readln(S);
  EnterPassword:=S;
end;



//*****************************************************************************
//* EnterBoolean
//*   Entering Boolean value with prompt
//*****************************************************************************
function EnterBoolean(Prompt: string): Boolean;
var
  S: string;
begin
  write('Enter '+Prompt+': ');
  Readln(S);
  S:=UpperCase(S);
  if S='TRUE'then
    EnterBoolean:=True
  else
    EnterBoolean:=False;
end;



//*****************************************************************************
//* EnterInteger
//*   Entering Integer value with prompt
//*****************************************************************************
function EnterInteger(Prompt: string): Integer;
var
  S: string;
begin
  write('Enter '+Prompt+': ');
  Readln(S);
  try
    EnterInteger:=StrToInt(S);
  except
    EnterInteger:=0;
  end;
end;



//*****************************************************************************
//* EnterString
//*   Entering String value with prompt
//*****************************************************************************
function EnterString(Prompt: string): string;
var
  S: string;
begin
  write('Enter '+Prompt+': ');
  Readln(S);
  EnterString:=S;
end;



//*****************************************************************************
//* GetNumberOfMessagesInPOPServer
//*
//*****************************************************************************
function GetNumberOfMessagesInPOPServer(Host: string; Port: Integer; UserID: string; Password: string): Integer;
var
  POP3: TIdPOP3;
begin
  POP3:=TIdPOP3.Create(Nil);
  POP3.Host:=Host;
  POP3.Port:=Port;
  POP3.Username:=UserID;
  POP3.Password:=Password;
  try
    POP3.Connect;
  except
    POP3.Free;
    Result:=-1;
    Exit;
  end;
  Result:=POP3.CheckMessages;
  POP3.Disconnect;
  POP3.Free;
end;



//*****************************************************************************
//* SaveLogToFile
//*   Save logging message into the log file
//*****************************************************************************
function SaveLogToFile(Msg: string): Boolean;
var
  FullPath:    string;
  LogFile:     Text;
  LogName:     string;
  LogDateTime: TDateTime;
  Errored:     Boolean;
begin
  LogDateTime:=Now;
  LogName:=FormatDateTime('yyyymmdd', LogDateTime)+'.txt';
  SaveLogToFile:=True;
  FullPath:=Param.LogDir+PathDelimiter+LogName;
  Assign(LogFile, FullPath);
  Errored:=False;
  if FileExists(FullPath) then
    Append(LogFile)
  else
  begin
    try
      Rewrite(LogFile);
    except
      Errored:=True;
    end;
    if Errored=False then
      Writeln(LogFile, '# MailUtl log file');
  end;

  if Errored then
    SaveLogToFile:=False
  else
  begin
    if Msg='' then
      Writeln(LogFile)
    else
      Writeln(LogFile, FormatDateTime('mmm/dd/yyyy hh:nn:ss ', LogDateTime)+Msg);
    CloseFile(LogFile);
  end;
end;



//*****************************************************************************
//* GetFieldValue
//*   finds designated field as "Field" constant,
//*   returns field value.
//*****************************************************************************
function GetFieldValue(S: string; Field: string): string;
var
  Index: Integer;
  Count: Integer;
begin
  Result:='';
  if (pos(UpperCase(Field), UpperCase(S) )=1) and (Length(S)>Length(Field)) then
  begin
    Index:=Length(Field)+1;
    Count:=Length(S)-Length(Field);
    Result:=Trim(Copy(S, Index, Count))
  end;
end;



//*****************************************************************************
//* GetFieldValueInList
//*   finds designated field as "Field" constant in the stringlist,
//*   returns field value.
//*   Note: If two or more fields are found, this function returns
//*         all mutched fields.  (Ex: To:, Cc:, From: or Received: field)
//*****************************************************************************
function GetFieldValueInList(Headers: TStringList; Field: string): string;
var
  FieldValue: string;
  Finish:     Boolean;
  I:          Integer;
  S:          string;
begin
  Result:='';
  Finish:=False;
  I:=0;
  while (I<Headers.Count)and(Finish=False) do
  begin
    S:=Headers[I];
    FieldValue:=GetFieldValue(S, Field);
    if FieldValue<>'' then
    begin
      if Result<>'' then
        Result:=Result+' , ';
      Result:=Result+FieldValue;
      Finish:=True;
      S:=Uppercase(Field);
      if (S='FROM:')or(S='TO:')or(S='CC:')or(S='REPLY-TO:')or(S='RECEIVED:') then
        Finish:=False;
    end;
    inc(I);
  end;
end;




//*****************************************************************************
//* ArrangeHeaders
//*   converts headers to Japanese,
//*   combine header lines which consist of several rows (ex: "Received:"
//*   header) into one line.
//*****************************************************************************
procedure ArrangeHeaders(AMsg: TIdMessage);
var
  Count:           Integer;
  FirstChar:       char;
  I:               Integer;
  LastField:       string;
  NumberOfHeaders: Integer;
  UpperStr:        string;
  TmpStr:          string;
begin
  NumberOfHeaders:=AMsg.Headers.Count;
  LastField:='';
  I:=0; Count:=0;
  repeat
    TmpStr:=AMsg.Headers[I];
    {$IFDEF Japanese}
    TmpStr:=Jap.MIMEHeaderDecode(TmpStr);
    TmpStr:=Jap.JisToSJis(TmpStr);
    {$ENDIF Japanese}
    UpperStr:=UpperCase(TmpStr);
    FirstChar:=TmpStr[1];
    if (FirstChar=' ')or(FirstChar=char($09)) then
    begin
      if LastField='SUBJECT:' then
        AMsg.Headers[I-1]:=AMsg.Headers[I-1]+Trim(TmpStr)
      else
        AMsg.Headers[I-1]:=AMsg.Headers[I-1]+' '+Trim(TmpStr);
      AMsg.Headers.Delete(I);
    end
    else
    begin
      AMsg.Headers[I]:=TmpStr;
      if pos('SUBJECT:', UpperStr)=1 then
        LastField:='SUBJECT:'
      else
        LastField:='';
      inc(I);
    end;
    inc(Count);
  until Count>=NumberOfHeaders;
  AMsg.From.Text:=GetFieldValueInList(AMsg.Headers, 'From:');
  AMsg.Subject:=GetFieldValueInList(AMsg.Headers, 'Subject:');
  AMsg.ReplyTo.EMailAddresses:=GetFieldValueInList(AMsg.Headers, 'Reply-To:');
  AMsg.Recipients.EMailAddresses:=GetFieldValueInList(AMsg.Headers, 'To:');
end;






//*****************************************************************************
//* InitParam
//*   Initialize parameters
//*****************************************************************************
procedure InitParam(var Param: TMyParam);
begin
  with Param do
  begin
    POP3_Host:='';
    POP3_Port:=110;
    POP3_UserID:='';
    POP3_Password:='';
    LogDir:='c:\';
    Mailer:='MailUtl ver. '+Version;
    TemplateDir:='c:\';
    InhibitAction:=False;
    HTML_Lang:='';
    HTML_Charset:='';
  end;
end;



//*****************************************************************************
//* ReadFromIniFile
//*   Read parameters from the .INI file
//*****************************************************************************
function ReadFromIniFile(IniName: string): boolean;
var
  IniFile: TIniFile;
begin
  ReadFromIniFile:=True;
  IniFile := TIniFile.Create(IniName);
  try
    Param.IniName:=IniName;
    Param.LogDir:=IniFile.ReadString('Dir', 'Log', Param.LogDir);
    Param.TemplateDir:=IniFile.ReadString('Dir', 'Template', Param.TemplateDir);
    Param.POP3_Host:=IniFile.ReadString('POP3', 'Host', Param.POP3_Host);
    Param.POP3_Port:=IniFile.ReadInteger('POP3', 'Port', Param.POP3_Port);
    Param.POP3_UserID:=IniFile.ReadString('POP3', 'UserID', Param.POP3_UserID);
    Param.HTML_Lang:=IniFile.ReadString('HTML', 'Lang', Param.HTML_Lang);
    Param.HTML_Charset:=IniFile.ReadString('HTML', 'Charset', Param.HTML_Charset);
  except
    ReadFromIniFile:=False;
  end;
  IniFile.Free;
end;

function SaveToIniFile(IniName: string): boolean;
var
  IniFile: TIniFile;
begin
  SaveToIniFile:=True;
  IniFile := TIniFile.Create(IniName);
  try
    IniFile.WriteString('General', 'Program', 'MailUtl');
    IniFile.WriteString('General', 'Version', Version);
    IniFile.WriteString('General', 'Copyright', 'Copyright (C) 2002 Fumio KAWAMATA');
    IniFile.WriteString('Dir', 'Log', Param.LogDir);
    IniFile.WriteString('Dir', 'Template', Param.TemplateDir);
    IniFile.WriteString('POP3', 'Host', Param.POP3_Host);
    IniFile.WriteInteger('POP3', 'Port', Param.POP3_Port);
    IniFile.WriteString('POP3', 'UserID', Param.POP3_UserID);
    IniFile.WriteString('HTML', 'Lang', Param.HTML_Lang);
    IniFile.WriteString('HTML', 'Charset', Param.HTML_Charset);
  except
    SaveToIniFile:=False;
  end;
  IniFile.Free;
end;



//*****************************************************************************
//* WriteUsage
//*   Write usage to console
//*****************************************************************************
procedure WriteUsage;
var
  StdOut: text;
begin
  Assign(StdOut, '');
  Rewrite(StdOut);
  Writeln(StdOut, 'A Mail Utility "MAILUTL" Version '+Version);
  Writeln(StdOut);
  Writeln(StdOut, 'This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.');
  Writeln(StdOut);
  Writeln(StdOut, 'This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.');
  Writeln(StdOut);
  Writeln(StdOut, 'You should have received a copy of the GNU General Public License along with this program.');
  Writeln(StdOut);
  Writeln(StdOut, 'MAILUTL  [-IIniFile]  [-PPOP3_Pass]  [-pPOP3_Pass2]  [-SSMTP_Pass]  [-RRuleFile]  [-C]  [-H]  [-B]');
  Writeln(StdOut, '   -C: Conversational command entry');
  Writeln(StdOut, '   -H: Show help message');
  Writeln(StdOut, '   -B: Inhibit Action (Ignore Action)');
  Writeln(StdOut);
  CloseFile(StdOut);
end;



//*****************************************************************************
//* SetParam
//*   Set parameters interactively
//*     Returns  True: Cintinue the program
//*              False: Abort the program
//*****************************************************************************
function SetParam: Boolean;
var
  Cmd:      string;
  ExitFlag: Boolean;
procedure Menu;
begin
  Writeln;
  Writeln('***** Menu *****');
  Writeln;
  Writeln(' 1. POP3 Host: ', Param.POP3_Host);
  Writeln(' 2. POP3 Port (Default is 110): ', Param.POP3_Port);
  Writeln(' 3. POP3 User ID: ', Param.POP3_UserID);
  Writeln(' 4. Log directory: ', Param.LogDir);
  Writeln(' 5. INI filename: ', Param.IniName);
  Writeln(' 6. HTML Lang: ', Param.HTML_Lang);
  Writeln(' 7. HTML Charset: ', Param.HTML_Charset);
  Writeln;
  Write('Command [1]-[7],  [S]ave,  [Q]uit without save,  [R]un ==> ');
  Readln(Cmd);
  if Cmd='5' then
    Param.IniName:=EnterString('INI filename')
  else
  if Cmd='4' then
    Param.LogDir:=EnterString('Log directory')
  else
  if Cmd='1' then
    Param.POP3_Host:=EnterString('POP3 Host')
  else
  if Cmd='2' then
    Param.POP3_Port:=EnterInteger('POP3 Port')
  else
  if Cmd='3' then
    Param.POP3_UserID:=EnterString('POP3 User ID')
  else
  if Cmd='6' then
    Param.HTML_Lang:=EnterString('HTML Lang')
  else
  if Cmd='7' then
    Param.HTML_Charset:=EnterString('HTML Charset')
  else
  if Cmd='Q' then
  begin
    ExitFlag:=True;
    SetParam:=False;
  end
  else
  if Cmd='R' then
    ExitFlag:=True
  else
  if Cmd='S' then
  begin
    if SaveToIniFile(Param.IniName)=False then
    begin
      Writeln('The program could not save to "'+Param.IniName+'".');
      Write('Hit [Enter] key to retry.  ');
      Readln;
    end;
  end
  else
  begin
    write('It is undefined command.  Hit [Enter] key to proceed.');
    Readln;
  end;
  Writeln;
  Writeln;
end;
begin
  SetParam:=True;
  ExitFlag:=False;
  while ExitFlag=False do
  begin
   Menu;
  end;
end;



//*****************************************************************************
//* SplitBeforeDelimiter
//*   Split string by delimiter
//*****************************************************************************
function SplitBeforeDelimiter(var S: string; Delimiter: Char): string;
var
  TabPos: Integer;
begin
  TabPos:=Pos(Delimiter, S);
  if TabPos=0 then
  begin
    Result:=Trim(S);
    S:='';
  end
  else
  if TabPos=1 then
  begin
    Result:='';
    Delete(S, 1, 1);
  end
  else
  begin
    Result:=Trim(Copy(S, 1, TabPos-1));
    Delete(S, 1, TabPos);
  end;
end;




//*****************************************************************************
//* ReadRuleFromFile
//*   Read rule file from the file
//*****************************************************************************
function ReadRuleFromFile(RuleName: string): Boolean;
var
  F:         Text;
  S:         string;
  SearchStr: string;
  AFieldStr: string;
  FieldsStr: string;
  ActionStr: string;
begin
  ReadRuleFromFile:=True;
  try
    Assign(F, RuleName);
    Reset(F);
  except
    ReadRuleFromFile:=False;
    Exit;
  end;
  while not(eof(F)) do
  begin
    Readln(F, S);
    S:=Trim(S);
    if (Pos('#', S)<>1)and(S<>'') then
    begin
      SearchStr:=SplitBeforeDelimiter(S, Chr($09));
      if SearchStr='' then
      begin
        CloseFile(F);
        Exit;
      end;

      FieldsStr:=SplitBeforeDelimiter(S, Chr($09));
      if FieldsStr='' then
      begin
        FieldsStr:='FROM:'; //* Default value
      end;

      ActionStr:=SplitBeforeDelimiter(S, Chr($09));
      if ActionStr='' then
      begin
        ActionStr:='DEL'; //* Default value
      end;

      while FieldsStr<>'' do
      begin
        AFieldStr:=SplitBeforeDelimiter(FieldsStr, ',');
        if (SearchStr<>'')and(AFieldStr<>'')and(ActionStr<>'') then
        begin
          Field1List.Add(SearchStr);
          Field2List.Add(AFieldStr);
          Field3List.Add(ActionStr);
        end;
      end;
    end;
  end;
  CloseFile(F);
end;



//*****************************************************************************
//* ExecuteOption
//*   Execute optional command
//*****************************************************************************
function ExecuteOption: Integer;
var
  I: Integer;
  S: string;
begin
  ExecuteOption:=0;
  for I:=1 to ParamCount do
  begin
    S:=ParamStr(I);
    if pos('-B', S)=1 then
      Param.InhibitAction:=True
    else
    if pos('-I', S)=1 then
    begin
      Delete(S,1,2);
      if S<>'' then
        ReadFromIniFile(S);
    end
    else
    if pos('-P',S)=1 then
    begin
      Delete(S,1,2);
      if S<>'' then
        Param.POP3_Password:=S;
    end
    else
    if pos('-R',S)=1 then
    begin
      Delete(S,1,2);
      if S<>'' then
      begin
        if ReadRuleFromFile(S)=False then
          ExecuteOption:=3;
      end;
    end
    else
    if pos('-C',S)=1 then
    begin
      if SetParam=False then
        ExecuteOption:=2;
    end
    else
    if pos('-H',S)=1 then
    begin
      WriteUsage;
      ExecuteOption:=1;
    end;
  end;
end;




//*****************************************************************************
//* SetShortMonthNames
//*****************************************************************************
procedure SetShortMonthNames;
const
  ShortMonthNamesArray: array[1..12] of string = ('Jan', 'Feb', 'Mar', 'Apr',
    'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec');
var
  I: Integer;
begin
  for I:=1 to 12 do
    ShortMonthNames[I]:=ShortMonthNamesArray[I];
end;



//*****************************************************************************
//* CheckField
//*****************************************************************************
function CheckField(Headers: TStringList; Field: string; S: string): Boolean;
var
  FieldValue:    string;
  S2:            string;
  NegativeLogic: Boolean;
  RegularExpression: Boolean;
begin
  RegularExpression:=False;
  NegativeLogic:=False;
  Result:=False;
  S2:=S;
  if pos('!', S2)=1 then
  begin
    NegativeLogic:=True;
    Delete(S2, 1,1);
    S2:=Trim(S2);
  end;
  if pos('\', S2)=1 then
  begin
    Delete(S2, 1, 1);
    RegularExpression:=True;
  end;

  if S2<>'' then
  begin
    FieldValue:=GetFieldValueInList(Headers, Field);
    if RegularExpression=True then
    begin
      Result:=brx.Match(S2, FieldValue);
    end
    else begin
      if pos(UpperCase(S2), UpperCase(FieldValue))>0 then
        Result:=True;
    end;
    if NegativeLogic=True then
    begin
      Result:=not(Result);
    end;
  end;
end;



//*****************************************************************************
//* CheckRule
//*****************************************************************************
function CheckRule(RxMsg: TIdMessage; MsgSize: Integer): TMyEvaluated;
var
  Finish:         Boolean;
  RuleIndex:      Integer;
  Matched:        Boolean;
  MatchedReason:  string;
  MisDateTime:    TDateTime;
  AllowedMisDate: Integer;
  S:              string;
  Size:           Integer;
begin
  Result.Action:=actNop;
  Result.Reason:='';
  RuleIndex:=0;
  Finish:=False;
  while (RuleIndex<Field1List.Count)and(Finish=False) do
  begin
    Matched:=False;

    //* Message size
    if (Pos('>',Field1List[RuleIndex])=1)and(UpperCase(Field2List[RuleIndex])='MSGSIZE') then
    begin
      S:=Field1List[RuleIndex];
      Delete(S, 1, 1);
      S:=Trim(S);
      if S<>'' then
      begin
        Size:=MyStrToInt(S);
        if MsgSize>Size then
        begin
          Matched:=True;
          MatchedReason:='Message size ('+IntToStr(MsgSize)+')>'+IntToStr(Size);
        end;
      end;
    end;

    //* Future Date
    if (Pos('>',Field1List[RuleIndex])=1)and(UpperCase(Field2List[RuleIndex])='FUTUREDATE') then
    begin
      S:=Field1List[RuleIndex];
      Delete(S, 1, 1);
      S:=Trim(S);
      if S<>'' then
      begin
        AllowedMisDate:=MyStrToInt(S);
        MisDateTime:=RxMsg.Date-Now;
        if MisDateTime>AllowedMisDate then
        begin
          Matched:=True;
          MatchedReason:='Misdated ('+DateTimeToStr(RxMsg.Date)+' is future date over '+IntToStr(AllowedMisDate)+' days)';
        end;
      end;
    end;


    //* Past Date
    if (Pos('>',Field1List[RuleIndex])=1)and(UpperCase(Field2List[RuleIndex])='PASTDATE') then
    begin
      S:=Field1List[RuleIndex];
      Delete(S, 1, 1);
      S:=Trim(S);
      if S<>'' then
      begin
        AllowedMisDate:=MyStrToInt(S);
        MisDateTime:=Now-RxMsg.Date;
        if MisDateTime>AllowedMisDate then
        begin
          Matched:=True;
          MatchedReason:='Misdated ('+DateTimeToStr(RxMsg.Date)+' is past date over '+IntToStr(AllowedMisDate)+' days)';
        end;
      end;
    end;


    if CheckField(RxMsg.Headers, Field2List[RuleIndex], Field1List[RuleIndex])=True then
    begin
      Matched:=True;
      MatchedReason:='"'+Field1List[RuleIndex]+'" in "'+Field2List[RuleIndex]+'"';
    end;

    S:=UpperCase(Field3List[RuleIndex]);

    if Matched=True then
    begin
      if Result.Reason<>'' then
      begin
        Result.Reason:=Result.Reason+' and ';
      end;
      Result.Reason:=Result.Reason+MatchedReason;
      if S<>'CONTINUE' then
      begin
        if S='KEEP' then
        begin
          Result.Action:=actKeep
        end
        else
        if S='DEL' then
        begin
          Result.Action:=actDel
        end
        else
        begin
          Result.Action:=actNop;
        end;
        Finish:=True;
      end;
    end
    else
    begin //* UnMatched
      if S='CONTINUE' then
      begin //* Skip all 'CONTINUE' actions.
        while (RuleIndex<(Field1List.Count-1))and(Uppercase(Field3List[RuleIndex])='CONTINUE') do
        begin
          Inc(RuleIndex);
        end;
      end;
      Result.Reason:='';
    end;

    if Finish=False then
    begin
      if RuleIndex<Field1List.Count then
      begin
        Inc(RuleIndex);
      end;
    end;
  end;
end;



//*****************************************************************************
//* ExtractNumberPart
//* This function is called from GetNewFullPath function.
//*****************************************************************************
function ExtractNumberPart(InStr: string): string;
var
  PositionOf_: Integer;
  WorkStr: string;
begin
  WorkStr:=ExtractFileName(InStr);
  WorkStr:=ChangeFileExt(WorkStr, '');
  PositionOf_:=LastDelimiter('_', WorkStr);
  if PositionOf_>0
    then begin
      Delete(WorkStr, 1, PositionOf_);
      Result:=WorkStr;
    end
  else
    Result:='';
end;



//*****************************************************************************
//* GetNewFullPath
//* This function requires ExtractNumberPart function.
//*****************************************************************************
function GetNewFullPath(FullPath: string): string;
var
  FileName:           string;
  FileNameWithoutExt: string;
  FilePath:           string;
  I:                  Integer;
  J:                  Integer;
  NumberList:         TStrings;
  NumberPart:         string;
  NewNumber:          Integer;
  SearchRec:          TSearchRec;
  TmpStr:             string;
begin
  Result:=FullPath;
  FilePath:=ExtractFilePath(FullPath);
  FileName:=ExtractFileName(FullPath);
  FileNameWithoutExt:=ChangeFileExt(FileName, '');

  (* Create and Clear NumberList *)
  NumberList:=TStringList.Create;
  NumberList.Clear;

  (* Find first match *)
  if FindFirst(FilePath+FileNameWithoutExt+'*.txt', faAnyFile, SearchRec)=0
    then begin
      NumberPart:=ExtractNumberPart(SearchRec.Name);
      if NumberPart<>''
        then begin
          NumberList.Add(NumberPart);
        end;
    end;

  (* Get number part of the other matched filename *)
  while FindNext(SearchRec)=0 do
    begin
      NumberPart:=ExtractNumberPart(SearchRec.Name);
      if NumberPart<>''
      then begin
        NumberList.Add(NumberPart);
      end;
    end;

  (* Sort the NumberList *)
  if NumberList.Count>=2
  then begin
    for I:=0 to NumberList.Count-2 do
      for J:=I+1 to NumberList.Count-1 do
        begin
          if StrToInt(NumberList[I])>StrToInt(NumberList[J])
            then begin
              TmpStr:=NumberList[J];
              NumberLIst[J]:=NumberList[I];
              NumberLIst[I]:=TmpStr
            end;
        end;
    end;
  if NumberList.Count=0
    then begin
      NewNumber:=2;
    end
  else begin
    NewNumber:=StrToInt(NumberList[NumberList.Count-1])+1;
  end;

  Result:=FilePath+FileNameWithoutExt+'_'+IntToStr(NewNumber)+'.txt';

  NumberList.Free;
end;



//*****************************************************************************
//* SaveHeadersToFile
//*****************************************************************************
function SaveHeadersToFile(Param: TMyParam; RxMsg: TIdMessage): string;
var
  I:        Integer;
  F:        Text;
  FileName: string;
  FullPath: string;
begin
  FileName:=FormatDateTime('yyyymmdd-hhnnss', RxMsg.Date)+'.txt';
  Result:=FileName;
  FullPath:=Param.LogDir+PathDelimiter+FileName;
  if FileExists(FullPath)
    then begin
      FullPath:=GetNewFullPath(FullPath);
      Result:=ExtractFileName(FullPath);
    end;

  try
    Assign(F, FullPath);
    Rewrite(F);
    for I:=0 to RxMsg.Headers.Count-1 do
      begin
        Writeln(F, RxMsg.Headers[I]);
      end;
    CloseFile(F);
  except
    Result:='';
    Exit;
  end;
end;



//*****************************************************************************
//* EncodeForHTML
//*****************************************************************************
function EncodeForHTML(S: string): string;
var
  I:   Integer;
  WS:  WideString;
  Tmp: string;
begin
  WS:=S;
  Result:='';
  if WS='' then
    Exit;
  for I:=1 to Length(WS) do
  begin
    Tmp:=copy(WS, I, 1);
    if Tmp='<' then
      Tmp:='&lt;'
    else
    if Tmp='>' then
      Tmp:='&gt;'
    else
    if Tmp='&' then
      Tmp:='&amp;'
    else
    if Tmp='"' then
      Tmp:='&quot;'
    ;
    Result:=Result+Tmp;
  end;
end;



//*****************************************************************************
//* Wait
//*****************************************************************************
procedure Wait(WaitTime: Integer);
var
  LastTick: Cardinal;
begin
  LastTick:=GetTickCount;
  repeat
  until GetTickCount>=(LastTick+Cardinal(WaitTime));
end;



procedure _Main; begin end;
//*****************************************************************************
//* Main routine
//* HaltCode:
//*   0: Normally finished.
//*   1: Could not read .INI file.
//*   2: Aborted after displaying the usage.
//*   3: Aborted by user command.
//*   4: Rule file error.
//*   5: Could not connect to the POP server.
//*   6: No message in the POP server.
//*   7: Could not create the HTML file.
//*****************************************************************************
var
  BGColor:          string;
  CheckRuleResult:  TMyEvaluated;
  HTMLFile:         Text;
  HTMLName:         string;
  I:                Integer;
  MsgSize:          Integer;
  CurrentDateTime:  TDateTime;
  NumberOfMessages: Integer;
  POP3:             TIdPOP3;
  Result:           Integer;
  RxMsg:            TIdMessage;
  S:                string;
  SavedFileName:    string;
  Title:            string;
  TxMsg:            TIdMessage;
begin
  Writeln('MailUtl '+Version+'  Copyright (C) 2002-2005 Fumio KAWAMATA');
  SetShortMonthNames;
  InitParam(Param);
  Param.POP3_Password:=GetEnvironmentVariable('POP3_USERPASS');
  Param.IniName:=ChangeFileExt(ParamStr(0), '.ini');
  if FileExists(Param.IniName) then
  begin
    try
      ReadFromIniFile(Param.IniName);
    except
      Writeln('Could not read "'+Param.IniName+'".');
      Halt(1);
    end;
  end;
  Field1List:=TStringList.Create;
  Field2List:=TStringList.Create;
  Field3List:=TStringList.Create;
  Result:=ExecuteOption;
  //* ExecuteOption returns
  //* 1: Help, 2: Quit by User Command, 3: Rule file error
  if Result>0 then
  begin
    Field3List.Free;
    Field2List.Free;
    Field1List.Free;
    case Result of
      1: S:='Showed help message by user command.';
      2: S:='Quit by user command.';
      3: S:='Rule file error.';
      else S:='Halt Code: '+IntToStr(Result+1)
    end;
    Writeln(S);
    Halt(Result+1);
  end;

  if Param.POP3_Password='' then
    Param.POP3_Password:=EnterPassword('POP3');

  SaveLogToFile('');
  SaveLogToFile('Program started.');

  S:=Param.POP3_Host+' for chekking number of messages.';
  Writeln('Connecting to '+S);
  SaveLogToFile('Connect to '+S);
  NumberOfMessages:=GetNumberOfMessagesInPOPServer(Param.POP3_Host, Param.POP3_Port, Param.POP3_UserID, Param.POP3_Password);
  if NumberOfMessages<0 then
  begin
    S:='Failed connection.  Check POP server, user id or password.  Program aborted.';
    Writeln(S);
    SaveLogToFile(S);
    Halt(5);
  end;

  if NumberOfMessages=0 then
  begin
    S:='No message in the POP server.  Program aborted.';
    Writeln(S);
    SaveLogToFile(S);
    Halt(6);
  end;

  if NumberOfMessages>1 then
    S:='There were '
  else
    S:='There was ';
  S:=S+IntToStr(NumberOfMessages);
  if NumberOfMessages>1 then
    S:=S+' messages.'
  else
    S:=S+' message.';
  Writeln(S);
  SaveLogToFile(S);


  {$IFDEF Japanese}
  Jap:=TPZConv.Create(Nil);
  {$ENDIF Japanese}


  POP3:=TIdPOP3.Create(Nil);
  POP3.Host:=Param.POP3_Host;
  POP3.Port:=Param.POP3_Port;
  POP3.UserName:=Param.POP3_UserID;
  POP3.Password:=Param.POP3_Password;
  S:='Host: '+POP3.Host+' , Port:'+IntToStr(POP3.Port);
  Writeln('Connecting to '+S);
  SaveLogToFile('Connect to '+S);
  try
    POP3.Connect;
  except
    S:='Failed connection.  Check POP server, user id or password.  Program aborted.';
    Writeln(S);
    SaveLogToFile(S);
    POP3.Free;
    {$IFDEF Japanese}
    Jap.Free;
    {$ENDIF Japanese}
    Field3List.Free;
    Field2List.Free;
    Field1List.Free;
    Halt(5);
  end;

  CurrentDateTime:=Now;
  HTMLName:=Param.LogDir+PathDelimiter+FormatDateTime('yyyymmdd-hhnnss', CurrentDateTime)+'.html';
  Assign(HTMLFile, HTMLName);
  try
    Rewrite(HTMLFile);
  except
    S:='Could not create the HTML file.  Program aborted.';
    Writeln(S);
    SaveLogToFile(S);
    POP3.Disconnect;
    POP3.Free;
    {$IFDEF Japanese}
    Jap.Free;
    {$ENDIF Japanese}
    Field3List.Free;
    Field2List.Free;
    Field1List.Free;
    Halt(7);
  end;

  Title:='MailUtl status log  '+FormatDateTime('mmm/dd/yyyy  hh:nn:ss', CurrentDateTime);
  Writeln(HTMLFile, '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 3.2//EN">');
  //Writeln(HTMLFile, '<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01//EN">');
  Writeln(HTMLFile, '<HTML>');
  Writeln(HTMLFile, '<HEAD>');
  if Param.HTML_Charset<>'' then
    Writeln(HTMLFile, '<META HTTP-EQUIV="Content-Type" CONTENT="text/html; charset='
      +Param.HTML_Charset+'">');
  Writeln(HTMLFile, '<TITLE>'+Title+'</TITLE>');
  Writeln(HTMLFile, '</HEAD>');
  Writeln(HTMLFile, '<BODY>');
  Writeln(HTMLFile, '<H1>'+EncodeForHTML(Title)+'</H1>');


  S:='Connected to '+S;
  Writeln(S);
  SaveLogToFile(S);
  NumberOfMessages:=POP3.CheckMessages;
  S:='Number of ';
  if NumberOfMessages>1 then
    S:=S+'messages: '
  else
    S:=S+'message: ';
  S:=S+IntToStr(NumberOfMessages);
  Writeln(S);
  SaveLogToFile(S);
  if NumberOfMessages>0 then
  begin
    Writeln(HTMLFile, '<TABLE BORDER="1" CELLSPACING="0">');
    Writeln(HTMLFile, '<TR><TH>No.</TH><TH>Headers and Action</TH></TR>');
    RxMsg:=TIdMessage.Create(Nil);
    TxMsg:=TIdMessage.Create(Nil);
    for I:=1 to NumberOfMessages do
    begin
      S:=IntToStr(I)+'/'+IntToStr(NumberOfMessages);
      Writeln('No. '+S);
      SaveLogToFile(S);
      MsgSize:=POP3.RetrieveMsgSize(I);
      RxMsg.Clear;
      if (POP3.RetrieveHeader(I, RxMsg)=True) and (MsgSize>=0) then
      begin
        POP3.KeepAlive;
        ArrangeHeaders(RxMsg);

        CheckRuleResult:=CheckRule(RxMsg, MsgSize);

        case CheckRuleResult.Action of
          actKeep:   BGColor:='#E0FFE0';
          actDel:    BGColor:='#FFFFE0';
          actNop:    BGColor:='#FFFFFF';
        end;

        Writeln(HTMLFile, '<TR><TD BGCOLOR="'+BGColor+'">'+S+'</TD><TD BGCOLOR="'+BGColor+'">');
        Writeln(HTMLFile, '<P>');
        if RxMsg.Subject<>'' then
        begin
          S:='Subject: '+RxMsg.Subject;
          Writeln(S);
          SaveLogToFile(S);
          Writeln(HTMLFile, EncodeForHTML(S)+'<BR>');
        end;
        if RxMsg.From.Text<>'' then
        begin
          S:='From: '+RxMsg.From.Text;
          Writeln(S);
          SaveLogToFile(S);
          Writeln(HTMLFile, EncodeForHTML(S)+'<BR>');
        end;
        if RxMsg.Recipients.EMailAddresses<>'' then
        begin
          S:='To: '+RxMsg.Recipients.EMailAddresses;
          Writeln(S);
          SaveLogToFile(S);
          Writeln(HTMLFile, EncodeForHTML(S)+'<BR>');
        end;
        if RxMsg.CcList.EMailAddresses<>'' then
        begin
          S:='Cc: '+RxMsg.CcList.EMailAddresses;
          Writeln(S);
          SaveLogToFile(S);
          Writeln(HTMLFile, EncodeForHTML(S)+'<BR>');
        end;
        Writeln(HTMLFile, '</P>');

        if (CheckRuleResult.Action=actDel) then
        begin
          SavedFileName:=SaveHeadersToFile(Param, RxMsg);
          if CheckRuleResult.Action=actDel then
          begin
            S:='Action: Del,  Reason: '+CheckRuleResult.Reason;
            Writeln(S);
            SaveLogToFile(S);
            Writeln(HTMLFile, '<P>'+EncodeForHTML(S)+'</P>');
            Writeln(HTMLFile, '<P><A HREF="'+SavedFileName+'">Full Headers</A></P>');
          end;
          if Param.InhibitAction=False then
          begin
            //* Delete mail on POP server.
            POP3.Delete(I);
          end;
        end;

        case CheckRuleResult.Action of
          actKeep:
          begin
            S:='Action: Keep, Reason: '+CheckRuleResult.Reason;
            Writeln(S);
            SaveLogToFile(S);
            Writeln(HTMLFile, '<P>'+EncodeForHTML(S)+'</P>');
          end;
          actNop:
          begin
            S:='Action: Nop';
            Writeln(S);
            SaveLogToFile(S);
            Writeln(HTMLFile, '<P>'+S+'</P>');
          end;
        end;
        Writeln(HTMLFile, '</TD></TR>');
      end
      else
      begin
        SaveLogToFile('Could not retrieve header.');
      end;
    end;
    TxMsg.Free;
    RxMsg.Free;
    Writeln(HTMLFile, '</TABLE>');
  end;
  POP3.Disconnect;
  S:='Disconnected.';
  Writeln(S);
  SaveLogToFile(S);
  POP3.Free;
  Writeln(HTMLFile, '</BODY>');
  Writeln(HTMLFile, '</HTML>');
  CloseFile(HTMLFile);
  {$IFDEF Japanese}
  Jap.Free;
  {$ENDIF Japanese}
  Field3List.Free;
  Field2List.Free;
  Field1List.Free;
end.
